package VISD::Server; # $Id: Server.pm 123 2020-10-13 16:24:10Z abalama $
use strict;
use warnings FATAL => 'all';
use utf8;

=encoding utf8

=head1 NAME

VISD::Server - VISD web-server class

=head1 VERSION

Version 1.01

=head1 SYNOPSIS

    use VISD::Server;

=head1 DESCRIPTION

This module provides VISD web-server functionality (visd programm)

=head2 reload

The reload hook

=head2 startup

Mojo application startup method

=head1 SEE ALSO

L<Mojolicious>, L<VISD>

=head1 AUTHOR

Serż Minus (Sergey Lepenkov) L<http://www.serzik.com> E<lt>abalama@cpan.orgE<gt>

=head1 COPYRIGHT

Copyright (C) 1998-2020 D&D Corporation. All Rights Reserved

=head1 LICENSE

This program is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.

See C<LICENSE> file and L<https://dev.perl.org/licenses/>

=cut

use Mojo::Base 'Mojolicious';

use Mojo::File qw/path/;
use Mojo::Util qw/sha1_sum secure_compare/;

use CTK::Util qw/preparedir sharedir/;
use CTK::ConfGenUtil;
use CTKx;
use VISD;

use vars qw/ $VERSION /;
$VERSION = "1.01";

has ctk => sub { CTKx->instance->ctk };


# Mojo Routes (paths)
sub startup {
    my $self = shift;
    my $ctk = $self->ctk;

    # Set password for Cookies signing
    my $username = value($ctk->conf("username") // "");
    my $secret = value($ctk->conf("password") // "");
    $self->secrets([$secret]) if length($secret);

    # Server logging
    my $lvl = $ctk->debugmode() ? "debug" : $ctk->conf("loglevel") || "debug";
    if ($lvl =~ /^warn/) { $lvl = "warn" }
    elsif ($lvl =~ /^not/) { $lvl = "info" }
    elsif ($lvl =~ /^(crit|alert|emerg|except)/) { $lvl = "fatal" }
    else { $lvl = "debug" }
    $self->log->level($lvl)->path($ctk->logfile());

    # Switch to installable home directory
    $self->home(Mojo::Home->new($ctk->datadir()));

    # Get DocumentRoot and replace as public-path
    my $documentroot = value($ctk->conf("documentroot")) || path(sharedir(), $ctk->prefix, "www")->to_string();
    $self->static->paths()->[0] = $documentroot; #unshift @{$static->paths}, '/home/sri/themes/blue/public';

    # Hooks
    $self->hook(before_dispatch => sub {
        my $c = shift;

        # Set Server header
        $c->res->headers->server(sprintf("%s/%s (Mojolicious)", VISD::PROJECTNAME, VISD::VISD_VERSION));

        # Authentication
        if (length($username)) {
            my $req_uri = $c->req->url->to_abs();
            my $ui_username = $req_uri->username() // "";
            my $ui_secret = sha1_sum($req_uri->password() // time());

            # Check username and password
            return 1 if length($secret)
                and secure_compare($username, $ui_username)
                and secure_compare($secret, $ui_secret);

            # Require authentication
            $c->res->headers->www_authenticate('Basic realm="VISD Strict Zone"');
            $c->render(text => 'Authentication required!', status => 401);
        }
    });

    # Routes
    $self->routes->get('/')->to('alpha#root');
    $self->routes->get('/cetus-data')->to('alpha#cetus_data');
    $self->routes->get('/env')->to('alpha#env') if $ctk->debugmode();
    $self->routes->get('/test')->to('alpha#test') if $ctk->testmode();
    #$self->routes->get('/hello')->to('alpha#hello');

    return 1;
}

# Reload hook
sub reload {
    my $self = shift;

    $self->log->debug("Request for reload $$"); # $self->ctk->logdir()

    return 1; # 1 - ok; 0 - error :(
}

1;

# Controller
package VISD::Server::Alpha;
use Mojo::Base 'Mojolicious::Controller';

use Mojo::File qw/path/;
use Mojo::JSON qw/from_json/;
use Mojo::Asset::File;

use CTK::Util qw/sharedir/;
use CTK::ConfGenUtil;
use VISD;
use VISD::Tools;

use constant {
        CONTENT_TYPE_LD     => "application/json",
        CONTENT_TYPE_PNG    => "image/png",
        CONTENT_TYPE_SVG    => "image/svg+xml",
        CONTENT_TYPE_TXT    => "text/plain",
    };

sub root {
    my $self = shift;
    #$self->render(text => 'Your Mojo is working! CTK: ' . $self->app->ctk->prefix());
    return $self->reply->static('index.html');
}
sub cetus_data {
    my $self = shift;
    my $ctk = $self->app->ctk();
    my $action = $self->req->url->query->param("action") || "";

    # Get last data from file
    my $ldf = path($ctk->datadir, VISD::LAST_DATA_FILE);
    return $self->render(json => { status => 0,
            error => sprintf("File \"%s\" not found", $ldf->to_string()),
        }, status => 404) unless -e $ldf->to_string();
    my $data = {};
    if (my $bindata = $ldf->slurp) {
        eval { $data = from_json($bindata) };
        return $self->render(json => { status => 0, error => $@ }, status => 500) if $@;
    } else {
        return $self->render(json => { status => 0,
            error => sprintf("Can't load file \"%s\"", $ldf->to_string()),
        }, status => 404) unless -e $ldf->to_string();
    }

    # Get ahosts from last data
    my $hostname = value($data, "hostname") // DEFAULT_HOSTNAME;
    my $ahosts = array($data, "ahosts");
    push @$ahosts, $hostname;
    my $ahosts_meta = hash($data, "ahosts_meta");
    my $net_interfaces = hash($data, "net/interfaces");
    my $fs_dev = hash($data, "fs/dev");
    $ahosts_meta->{$hostname} = {
        interfaces => [keys(%$net_interfaces)],
        devices => [keys(%$fs_dev)],
    };

    # Get host
    my $host = $self->req->url->query->param("host") || $hostname;
       $host =~ s/[^a-z0-9\-_.]//ig;
    return $self->render(json => { status => 0, error => "Host not specified!" }, status => 400) unless $host;
    return $self->render(json => { status => 0, error => "The \"$host\" host not found!" },
        status => 404) unless $ahosts_meta->{$host};


    #
    # Index: Returns last_data.json
    #


    unless ($action) {
        if ($host ne $hostname) {
            $ldf = path($ctk->datadir, lc($host), VISD::LAST_DATA_FILE);
            return $self->render(json => { status => 0,
                error => sprintf("File \"%s\" not found", $ldf->to_string()),
            }, status => 404) unless -e $ldf->to_string();
        }
        my $len = (-s $ldf->to_string()) || 0;
        return $self->render(json => { status => 0,
            error => sprintf("File \"%s\" is empty", $ldf->to_string()),
        }, status => 404) unless $len;
        $self->res->headers->content_type(CONTENT_TYPE_LD);
        return $self->reply->asset(Mojo::Asset::File->new(path => $ldf));
    }


    #
    # Graph
    #

    # Get group
    my $group = $self->req->url->query->param("group") || "sys";
    return $self->_error_graph(400 => 'The "group" query string parameter incorrect!')
        unless grep {$_ eq $group} @{ (VISD::Tools::AVAILABLE_GROUPS()) };

    # Get zooms
    my $zoom = $self->req->url->query->param("zoom") || "3h";
    return $self->_error_graph(400 => 'The "zoom" query string parameter incorrect!')
        unless grep {$_ eq $zoom} @{ (VISD::Tools::AVAILABLE_ZOOMS()) };

    # Get name
    my $name = $self->req->url->query->param("name") // "";
    if ($group eq 'net') {
        return $self->_error_graph(400 => 'The "name" query string parameter incorrect!')
            unless grep {$_ eq $name} @{ (array($ahosts_meta, $host, "interfaces")) };
    } elsif ($group =~ /^dev/) {
        return $self->_error_graph(400 => 'The "name" query string parameter incorrect!')
            unless grep {$_ eq $name} @{ (array($ahosts_meta, $host, "devices")) };
    }

    # Get RRD file name
    my $rrdf = path($ctk->datadir, lc($host), get_rrdfn($group, $name));

    # Plotting (trying)
    my ($rrdstatus, $rrdmsg);
    {
        local $@; # protect existing $@
        eval {
            ($rrdstatus, $rrdmsg) = rrd_graph(
                file => $rrdf->to_string,
                group => $group,
                zoom => $zoom,
            );
        };
        if ($@) {
            $rrdstatus = 0;
            $rrdmsg = $@;
        }
    }
    return $self->_error_graph(500 => sprintf("Can't plot RRD graph: %s", $rrdmsg)) unless $rrdstatus;

    # OUTPUT
    $self->res->headers->content_type(CONTENT_TYPE_PNG);
    return $self->render(data => $rrdmsg, format => 'png');

    ## Return JSON
    #return $self->render(json => {
    #    foo => "123",
    #    ldf => $ldf->to_string(),
    #    action => $action,
    #    rrdf => $rrdf->to_string(),
    #}); # , status => 403
}
sub env {
    my $self = shift;
    $self->res->headers->content_type(CONTENT_TYPE_TXT);
    return $self->render(text => $self->dumper(\%ENV));
}
sub test {
    my $self = shift;
    my $ctk = $self->app->ctk();

    #my $datadir = path(sharedstatedir(), VISD::PROJECTNAMEL)->to_string();

    my $documentroot = value($ctk->conf("documentroot")) || path(sharedir(), $ctk->prefix, "www")->to_string();
    #$self->app->log->debug("Catched!");

    $self->res->headers->content_type('text/plain');
    $self->render(data => VISD::_explain({
        #self        => $self, # Mojolicious::Routes
        datadir      => $ctk->datadir,
        documentroot => $documentroot,
        #listen_addr => $ctk->conf("listenaddr") || VISD::SERVER_LISTEN_ADDR,
        #listen_port => $ctk->conf("listenport") || VISD::SERVER_LISTEN_PORT,
        #UID         => $uid,
        #GID         => $gid,
        #ctk         => $ctk,
        #loglevel     => $ctk->conf("loglevel") || "debug",
        #admin        => sha1_sum("admin"),
        #req_username => $ctk->conf("username") // "",
        #req_password => $ctk->conf("password") // "",
        #homedir => $self->app->home(),
        #paths => $self->app->static->paths(),

    }));
}

#sub hello {
#    my $self = shift;
#    return $self->_error_graph(500 => "Oops!");
#}

sub _error_graph {
    my $self = shift;
    my $err = pop;
    my $code = pop || 500;

    $self->app->log->error($err);
    $self->res->headers->content_type(CONTENT_TYPE_SVG);

    return $self->render(data => <<SVG);
<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<svg xmlns="http://www.w3.org/2000/svg" id="error" version="1.1" width="640" height="260">
    <rect id="rect_error" width="100%" height="100%" style="fill: #CCCC88; fill-opacity:0.75" />
    <text id="text_title" x="25" y="50" style="font-weight: bold; font-size: 32; line-height:3.25; font-family: sans-serif; fill: #f24252">
      Error #$code
    </text>
    <text id="text_body" x="25" y="100" style="font-weight: normal; font-size: 14; line-height:1.25; font-family: sans-serif; fill: #f24252">
      $err
    </text>
</svg>
SVG
}

1;

__END__
